@@ -1,13 +1,18 @@
-#
-# imputil.py: import utilities
-#
+"""
+Import utilities
 
-### docco needed here and in Docs/ ...
+Exported classes:
+    ImportManager   Manage the import process
+
+    Importer        Base class for replacing standard import functions
+    BuiltinImporter Emulate the import mechanism for builtin and frozen modules
+
+    DynLoadSuffixImporter
+"""
 
 # note: avoid importing non-builtin modules
 import imp                      ### not available in JPython?
 import sys
-import strop
 import __builtin__
 
 # for the DirectoryImporter
@@ -78,7 +83,7 @@ class ImportManager:
     def _import_hook(self, fqname, globals=None, locals=None, fromlist=None):
         """Python calls this hook to locate and import a module."""
 
-        parts = strop.split(fqname, '.')
+        parts = fqname.split('.')
 
         # determine the context of this import
         parent = self._determine_import_context(globals)
@@ -126,6 +131,10 @@ class ImportManager:
         if importer:
             return importer._finish_import(top_module, parts[1:], fromlist)
 
+        # Grrr, some people "import os.path"
+        if len(parts) == 2 and hasattr(top_module, parts[1]):
+            return top_module
+
         # If the importer does not exist, then we have to bail. A missing
         # importer means that something else imported the module, and we have
         # no knowledge of how to get sub-modules out of the thing.
@@ -157,7 +166,7 @@ class ImportManager:
             assert globals is parent.__dict__
             return parent
 
-        i = strop.rfind(parent_fqname, '.')
+        i = parent_fqname.rfind('.')
 
         # a module outside of a package has no particular import context
         if i == -1:
@@ -291,7 +300,11 @@ class Importer:
             exec code in module.__dict__
 
         # fetch from sys.modules instead of returning module directly.
-        return sys.modules[fqname]
+        # also make module's __name__ agree with fqname, in case
+        # the "exec code in module.__dict__" played games on us.
+        module = sys.modules[fqname]
+        module.__name__ = fqname
+        return module
 
     def _load_tail(self, m, parts):
         """Import the rest of the modules, down from the top-level module.
@@ -612,7 +625,6 @@ def _test_revamp():
 # TODO
 #
 # from Finn Bock:
-#   remove use of "strop" -- not available in JPython
 #   type(sys) is not a module in JPython. what to use instead?
 #   imp.C_EXTENSION is not in JPython. same for get_suffixes and new_module
 #
@@ -638,7 +650,6 @@ def _test_revamp():
 #   need __path__ processing
 #   performance
 #   move chaining to a subclass [gjs: it's been nuked]
-#   avoid strop
 #   deinstall should be possible
 #   query mechanism needed: is a specific Importer installed?
 #   py/pyc/pyo piping hooks to filter/process these files
