@@ -8,7 +8,8 @@
  *   Version 1.5b1, heavily extended for ncurses by Oliver Andrich:
  *   Copyright 1996,1997 by Oliver Andrich, Koblenz, Germany.
  *
- *   Tidied for Python 1.6, and currently maintained by AMK (amk1@bigfoot.com)
+ *   Tidied for Python 1.6, and currently maintained by
+ *   <akuchlin@mems-exchange.org>.
  *
  *   Permission is hereby granted, free of charge, to any person obtaining
  *   a copy of this source file to use, copy, modify, merge, or publish it
@@ -31,7 +32,7 @@
  *   PERFORMANCE OF THIS SOFTWARE.
  */
 
-/* CVS: $Id: _cursesmodule.c,v 2.51.2.1 2002/02/08 21:32:41 gvanrossum Exp $ */
+/* CVS: $Id: _cursesmodule.c,v 2.61.6.2 2002/03/17 19:02:10 mwh Exp $ */
 
 /*
 
@@ -95,7 +96,7 @@ Form extension (ncurses and probably SYSV):
 
 /* Release Number */
 
-char *PyCursesVersion = "2.1";
+char *PyCursesVersion = "2.2";
 
 /* Includes */
 
@@ -106,6 +107,11 @@ char *PyCursesVersion = "2.1";
 #define STRICT_SYSV_CURSES      /* Don't use ncurses extensions */
 #endif
 
+#ifdef __hpux
+#define _XOPEN_SOURCE_EXTENDED
+#define STRICT_SYSV_CURSES
+#endif
+
 #define CURSES_MODULE
 #include "py_curses.h"
 
@@ -114,12 +120,11 @@ char *PyCursesVersion = "2.1";
     curses module in other ways.  So the code will just specify 
     explicit prototypes here. */
 extern int setupterm(char *,int,int *);
-#ifdef sgi
-extern char *tigetstr(char *);
-extern char *tparm(char *instring, ...);
+#ifdef __sgi
+#include <term.h>
 #endif
 
-#if defined(sgi) || defined(__sun__)
+#if !defined(HAVE_NCURSES_H) && (defined(sgi) || defined(__sun))
 #define STRICT_SYSV_CURSES       /* Don't use ncurses extensions */
 typedef chtype attr_t;           /* No attr_t type is available */
 #endif
@@ -189,8 +194,8 @@ PyCurses_ConvertToChtype(PyObject *obj, chtype *ch)
 {
   if (PyInt_Check(obj)) {
     *ch = (chtype) PyInt_AsLong(obj);
-  } else if(PyString_Check(obj) &
-	    (PyString_Size(obj) == 1)) {
+  } else if(PyString_Check(obj) 
+	    && (PyString_Size(obj) == 1)) {
     *ch = (chtype) *PyString_AsString(obj);
   } else {
     return 0;
@@ -355,7 +360,7 @@ static void
 PyCursesWindow_Dealloc(PyCursesWindowObject *wo)
 {
   if (wo->win != stdscr) delwin(wo->win);
-  PyMem_DEL(wo);
+  PyObject_DEL(wo);
 }
 
 /* Addch, Addstr, Addnstr */
@@ -389,7 +394,7 @@ PyCursesWindow_AddCh(PyCursesWindowObject *self, PyObject *args)
     use_xy = TRUE;
     break;
   default:
-    PyErr_SetString(PyExc_TypeError, "addch requires 1 or 4 arguments");
+    PyErr_SetString(PyExc_TypeError, "addch requires 1 to 4 arguments");
     return NULL;
   }
 
@@ -525,7 +530,7 @@ PyCursesWindow_Bkgd(PyCursesWindowObject *self, PyObject *args)
     return NULL;
   }
 
-  return PyCursesCheckERR(wbkgd(self->win, bkgd | A_NORMAL), "bkgd");
+  return PyCursesCheckERR(wbkgd(self->win, bkgd | attr), "bkgd");
 }
 
 static PyObject *
@@ -561,12 +566,32 @@ PyCursesWindow_BkgdSet(PyCursesWindowObject *self, PyObject *args)
 static PyObject *
 PyCursesWindow_Border(PyCursesWindowObject *self, PyObject *args)
 {
-  chtype ls, rs, ts, bs, tl, tr, bl, br;
-  ls = rs = ts = bs = tl = tr = bl = br = 0;
-  if (!PyArg_ParseTuple(args,"|llllllll;ls,rs,ts,bs,tl,tr,bl,br",
-                        &ls, &rs, &ts, &bs, &tl, &tr, &bl, &br))
+  PyObject *temp[8];
+  chtype ch[8];
+  int i;
+
+  /* Clear the array of parameters */
+  for(i=0; i<8; i++) {
+       temp[i] = NULL;
+       ch[i] = 0;
+  }    
+  
+  if (!PyArg_ParseTuple(args,"|OOOOOOOO;ls,rs,ts,bs,tl,tr,bl,br",
+                        &temp[0], &temp[1], &temp[2], &temp[3],
+                        &temp[4], &temp[5], &temp[6], &temp[7]))
     return NULL;
-  wborder(self->win, ls, rs, ts, bs, tl, tr, bl, br);
+
+  for(i=0; i<8; i++) {
+      if (temp[i] != NULL && !PyCurses_ConvertToChtype(temp[i], &ch[i])) {
+          PyErr_Format(PyExc_TypeError,
+                       "argument %i must be a ch or an int", i+1);
+          return NULL;
+      }
+  }
+  
+  wborder(self->win,
+          ch[0], ch[1], ch[2], ch[3],
+          ch[4], ch[5], ch[6], ch[7]);
   Py_INCREF(Py_None);
   return Py_None;
 }
@@ -585,6 +610,19 @@ PyCursesWindow_Box(PyCursesWindowObject *self, PyObject *args)
   return Py_None;
 }
 
+#if defined(HAVE_NCURSES_H) || defined(MVWDELCH_IS_EXPRESSION)
+#define py_mvwdelch mvwdelch
+#else
+int py_mvwdelch(WINDOW *w, int y, int x)
+{
+  mvwdelch(w,y,x);
+  /* On HP/UX, mvwdelch already returns. On other systems,
+     we may well run into this return statement. */
+  return 0;
+}
+#endif
+
+
 static PyObject *
 PyCursesWindow_DelCh(PyCursesWindowObject *self, PyObject *args)
 {
@@ -598,7 +636,7 @@ PyCursesWindow_DelCh(PyCursesWindowObject *self, PyObject *args)
   case 2:
     if (!PyArg_Parse(args,"(ii);y,x", &y, &x))
       return NULL;
-    rtn = mvwdelch(self->win,y,x);
+    rtn = py_mvwdelch(self->win,y,x);
     break;
   default:
     PyErr_SetString(PyExc_TypeError, "delch requires 0 or 2 arguments");
@@ -668,7 +706,7 @@ PyCursesWindow_EchoChar(PyCursesWindowObject *self, PyObject *args)
     return NULL;
   }
   
-#if !defined(__NetBSD__)
+#ifdef WINDOW_HAS_FLAGS
   if (self->win->_flags & _ISPAD)
     return PyCursesCheckERR(pechochar(self->win, ch | attr), 
 			    "echochar");
@@ -721,7 +759,7 @@ PyCursesWindow_GetCh(PyCursesWindowObject *self, PyObject *args)
     PyErr_SetString(PyExc_TypeError, "getch requires 0 or 2 arguments");
     return NULL;
   }
-  return PyInt_FromLong(rtn);
+  return PyInt_FromLong((long)rtn);
 }
 
 static PyObject *
@@ -835,8 +873,9 @@ PyCursesWindow_Hline(PyCursesWindowObject *self, PyObject *args)
 		     &y, &x, &temp, &n, &attr))
       return NULL;
     code = wmove(self->win, y, x);
+    break;
   default:
-    PyErr_SetString(PyExc_TypeError, "hline requires 2 or 5 arguments");
+    PyErr_SetString(PyExc_TypeError, "hline requires 2 to 5 arguments");
     return NULL;
   }
 
@@ -1073,7 +1112,7 @@ PyCursesWindow_NoOutRefresh(PyCursesWindowObject *self, PyObject *args)
   int pminrow,pmincol,sminrow,smincol,smaxrow,smaxcol;
   int rtn;
 
-#if defined(__NetBSD__)
+#ifndef WINDOW_HAS_FLAGS
   if (0) {
 #else
   if (self->win->_flags & _ISPAD) {
@@ -1215,7 +1254,7 @@ PyCursesWindow_Refresh(PyCursesWindowObject *self, PyObject *args)
   int pminrow,pmincol,sminrow,smincol,smaxrow,smaxcol;
   int rtn;
   
-#if defined(__NetBSD__)
+#ifndef WINDOW_HAS_FLAGS
   if (0) {
 #else
   if (self->win->_flags & _ISPAD) {
@@ -1283,7 +1322,7 @@ PyCursesWindow_SubWin(PyCursesWindowObject *self, PyObject *args)
   }
 
   /* printf("Subwin: %i %i %i %i   \n", nlines, ncols, begin_y, begin_x); */
-#if !defined(__NetBSD__)
+#ifdef WINDOW_HAS_FLAGS
   if (self->win->_flags & _ISPAD)
     win = subpad(self->win, nlines, ncols, begin_y, begin_x);
   else
@@ -1305,7 +1344,6 @@ PyCursesWindow_Scroll(PyCursesWindowObject *self, PyObject *args)
   switch(ARG_COUNT(args)) {
   case 0:
     return PyCursesCheckERR(scroll(self->win), "scroll");
-    break;
   case 1:
     if (!PyArg_Parse(args, "i;nlines", &nlines))
       return NULL;
@@ -1325,7 +1363,6 @@ PyCursesWindow_TouchLine(PyCursesWindowObject *self, PyObject *args)
     if (!PyArg_Parse(args,"(ii);start,count",&st,&cnt))
       return NULL;
     return PyCursesCheckERR(touchline(self->win,st,cnt), "touchline");
-    break;
   case 3:
     if (!PyArg_Parse(args, "(iii);start,count,val", &st, &cnt, &val))
       return NULL;
@@ -1363,8 +1400,9 @@ PyCursesWindow_Vline(PyCursesWindowObject *self, PyObject *args)
 		     &y, &x, &temp, &n, &attr))
       return NULL;
     code = wmove(self->win, y, x);
+    break;
   default:
-    PyErr_SetString(PyExc_TypeError, "vline requires 2 or 5 arguments");
+    PyErr_SetString(PyExc_TypeError, "vline requires 2 to 5 arguments");
     return NULL;
   }
 
@@ -1434,8 +1472,9 @@ static PyMethodDef PyCursesWindow_Methods[] = {
 	{"noutrefresh",     (PyCFunction)PyCursesWindow_NoOutRefresh},
         /* Backward compatibility alias -- remove in Python 2.1 */
 	{"nooutrefresh",    (PyCFunction)PyCursesWindow_NoOutRefresh},
-	{"overlay",       (PyCFunction)PyCursesWindow_Overlay, METH_VARARGS},
-	{"overwrite",     (PyCFunction)PyCursesWindow_Overwrite, METH_VARARGS},
+	{"overlay",         (PyCFunction)PyCursesWindow_Overlay, METH_VARARGS},
+	{"overwrite",       (PyCFunction)PyCursesWindow_Overwrite,
+         METH_VARARGS},
 	{"putwin",          (PyCFunction)PyCursesWindow_PutWin},
 	{"redrawln",        (PyCFunction)PyCursesWindow_RedrawLine},
 	{"redrawwin",       (PyCFunction)PyCursesWindow_redrawwin},
@@ -1458,7 +1497,7 @@ static PyMethodDef PyCursesWindow_Methods[] = {
 	{"touchwin",        (PyCFunction)PyCursesWindow_touchwin},
 	{"untouchwin",      (PyCFunction)PyCursesWindow_untouchwin},
 	{"vline",           (PyCFunction)PyCursesWindow_Vline},
-	{NULL,		        NULL}   /* sentinel */
+	{NULL,		    NULL}   /* sentinel */
 };
 
 static PyObject *
@@ -1472,7 +1511,7 @@ PyCursesWindow_GetAttr(PyCursesWindowObject *self, char *name)
 PyTypeObject PyCursesWindow_Type = {
 	PyObject_HEAD_INIT(NULL)
 	0,			/*ob_size*/
-	"curses window",	/*tp_name*/
+	"_curses.curses window",	/*tp_name*/
 	sizeof(PyCursesWindowObject),	/*tp_basicsize*/
 	0,			/*tp_itemsize*/
 	/* methods */
@@ -1615,7 +1654,7 @@ PyCurses_EraseChar(PyObject *self, PyObject *args)
 
   ch = erasechar();
 
-  return PyString_FromString(&ch);
+  return PyString_FromStringAndSize(&ch, 1);
 }
 
 static PyObject *
@@ -1811,6 +1850,11 @@ PyCurses_InitScr(PyObject *self, PyObject *args)
 	SetDictInt("ACS_HLINE",         (ACS_HLINE));
 	SetDictInt("ACS_VLINE",         (ACS_VLINE));
 	SetDictInt("ACS_PLUS",          (ACS_PLUS));
+#if !defined(__hpux) || defined(HAVE_NCURSES_H)
+        /* On HP/UX 11, these are of type cchar_t, which is not an
+           integral type. If this is a problem on more platforms, a
+           configure test should be added to determine whether ACS_S1
+           is of integral type. */
 	SetDictInt("ACS_S1",            (ACS_S1));
 	SetDictInt("ACS_S9",            (ACS_S9));
 	SetDictInt("ACS_DIAMOND",       (ACS_DIAMOND));
@@ -1825,6 +1869,7 @@ PyCurses_InitScr(PyObject *self, PyObject *args)
 	SetDictInt("ACS_BOARD",         (ACS_BOARD));
 	SetDictInt("ACS_LANTERN",       (ACS_LANTERN));
 	SetDictInt("ACS_BLOCK",         (ACS_BLOCK));
+#endif
 	SetDictInt("ACS_BSSB",          (ACS_ULCORNER));
 	SetDictInt("ACS_SSBB",          (ACS_LLCORNER));
 	SetDictInt("ACS_BBSS",          (ACS_URCORNER));
@@ -1967,7 +2012,7 @@ PyCurses_KillChar(PyObject *self, PyObject *args)
 
   ch = killchar();  
 
-  return PyString_FromString(&ch);  
+  return PyString_FromStringAndSize(&ch, 1);  
 }  
 
 static PyObject *
@@ -2051,7 +2096,7 @@ static PyObject *
 PyCurses_NewWindow(PyObject *self, PyObject *args)
 {
   WINDOW *win;
-  int nlines, ncols, begin_y, begin_x;
+  int nlines, ncols, begin_y=0, begin_x=0;
 
   PyCursesInitialised
 
@@ -2059,19 +2104,18 @@ PyCurses_NewWindow(PyObject *self, PyObject *args)
   case 2:
     if (!PyArg_Parse(args,"(ii);nlines,ncols",&nlines,&ncols))
       return NULL;
-    win = newpad(nlines, ncols);
     break;
   case 4:
     if (!PyArg_Parse(args, "(iiii);nlines,ncols,begin_y,begin_x",
 		   &nlines,&ncols,&begin_y,&begin_x))
       return NULL;
-    win = newwin(nlines,ncols,begin_y,begin_x);
     break;
   default:
     PyErr_SetString(PyExc_TypeError, "newwin requires 2 or 4 arguments");
     return NULL;
   }
 
+  win = newwin(nlines,ncols,begin_y,begin_x);
   if (win == NULL) {
     PyErr_SetString(PyCursesError, catchall_NULL);
     return NULL;
@@ -2266,6 +2310,15 @@ PyCurses_tparm(PyObject *self, PyObject *args)
 		return NULL;
 	}
 	
+#if defined(__hpux) || defined(_AIX)
+	/* tparm is declared with 10 arguments on a few platforms
+	   (HP-UX, AIX). If this proves to be a problem on other 
+	   platforms as well, perhaps an autoconf test should be 
+	   added to determine whether tparm can be called with a 
+	   variable number of arguments. Perhaps the other arguments 
+	   should be initialized in this case also. */
+	result = tparm(fmt,i1,i2,i3,i4,i5,i6,i7,i8,i9);
+#else
 	switch (PyTuple_GET_SIZE(args)) {
 	case 1:
 		result = tparm(fmt);
@@ -2298,7 +2351,7 @@ PyCurses_tparm(PyObject *self, PyObject *args)
 		result = tparm(fmt,i1,i2,i3,i4,i5,i6,i7,i8,i9);
 		break;
 	}
-
+#endif /* defined(__hpux) || defined(_AIX) */
 	return PyString_FromString(result);
 }
 
@@ -2448,7 +2501,8 @@ static PyMethodDef PyCurses_methods[] = {
   {"resetty",             (PyCFunction)PyCurses_resetty},
   {"savetty",             (PyCFunction)PyCurses_savetty},
   {"setsyx",              (PyCFunction)PyCurses_setsyx},
-  {"setupterm",           (PyCFunction)PyCurses_setupterm, METH_VARARGS|METH_KEYWORDS},
+  {"setupterm",           (PyCFunction)PyCurses_setupterm,
+   METH_VARARGS|METH_KEYWORDS},
   {"start_color",         (PyCFunction)PyCurses_Start_Color},
   {"termattrs",           (PyCFunction)PyCurses_termattrs},
   {"termname",            (PyCFunction)PyCurses_termname},
@@ -2460,7 +2514,7 @@ static PyMethodDef PyCurses_methods[] = {
   {"unctrl",              (PyCFunction)PyCurses_UnCtrl},
   {"ungetch",             (PyCFunction)PyCurses_UngetCh},
   {"use_env",             (PyCFunction)PyCurses_Use_Env},
-  {NULL,		NULL}		/* sentinel */
+  {NULL,		  NULL}		/* sentinel */
 };
 
 /* Initialization function for the module */
@@ -2490,6 +2544,7 @@ init_curses(void)
 	/* Add a CObject for the C API */
 	c_api_object = PyCObject_FromVoidPtr((void *)PyCurses_API, NULL);
 	PyDict_SetItemString(d, "_C_API", c_api_object);
+	Py_DECREF(c_api_object);
 
 	/* For exception curses.error */
 	PyCursesError = PyErr_NewException("_curses.error", NULL, NULL);
