@@ -2,7 +2,6 @@
 #include "Python.h"
 #include <sys/resource.h>
 #include <sys/time.h>
-#include <unistd.h>
 #include <string.h>
 #include <errno.h>
 
@@ -169,7 +168,8 @@ ins(PyObject *dict, char *name, int value)
 	/* errors will be checked by initresource() */
 }
 
-void initresource(void)
+DL_EXPORT(void)
+initresource(void)
 {
 	PyObject *m, *d;
 
