@@ -13,10 +13,6 @@
 #include "eval.h"
 #include "marshal.h"
 
-#ifdef HAVE_UNISTD_H
-#include <unistd.h>
-#endif
-
 #ifdef HAVE_SIGNAL_H
 #include <signal.h>
 #endif
@@ -63,6 +59,11 @@ int Py_NoSiteFlag; /* Suppress 'import site' */
 int Py_UseClassExceptionsFlag = 1; /* Needed by bltinmodule.c: deprecated */
 int Py_FrozenFlag; /* Needed by getpath.c */
 int Py_UnicodeFlag = 0; /* Needed by compile.c */
+int Py_IgnoreEnvironmentFlag; /* e.g. PYTHONPATH, PYTHONHOME */
+/* _XXX Py_QnewFlag should go away in 2.3.  It's true iff -Qnew is passed,
+  on the command line, and is used in 2.2 by ceval.c to make all "/" divisions
+  true divisions (which they will be in 2.3). */
+int _Py_QnewFlag = 0;
 
 static int initialized = 0;
 
@@ -86,6 +87,17 @@ Py_IsInitialized(void)
 
 */
 
+static int
+add_flag(int flag, const char *envs)
+{
+	int env = atoi(envs);
+	if (flag < env)
+		flag = env;
+	if (flag < 1)
+		flag = 1;
+	return flag;
+}
+
 void
 Py_Initialize(void)
 {
@@ -93,17 +105,18 @@ Py_Initialize(void)
 	PyThreadState *tstate;
 	PyObject *bimod, *sysmod;
 	char *p;
+	extern void _Py_ReadyTypes(void);
 
 	if (initialized)
 		return;
 	initialized = 1;
 	
-	if ((p = getenv("PYTHONDEBUG")) && *p != '\0')
-		Py_DebugFlag = Py_DebugFlag ? Py_DebugFlag : 1;
-	if ((p = getenv("PYTHONVERBOSE")) && *p != '\0')
-		Py_VerboseFlag = Py_VerboseFlag ? Py_VerboseFlag : 1;
-	if ((p = getenv("PYTHONOPTIMIZE")) && *p != '\0')
-		Py_OptimizeFlag = Py_OptimizeFlag ? Py_OptimizeFlag : 1;
+	if ((p = Py_GETENV("PYTHONDEBUG")) && *p != '\0')
+		Py_DebugFlag = add_flag(Py_DebugFlag, p);
+	if ((p = Py_GETENV("PYTHONVERBOSE")) && *p != '\0')
+		Py_VerboseFlag = add_flag(Py_VerboseFlag, p);
+	if ((p = Py_GETENV("PYTHONOPTIMIZE")) && *p != '\0')
+		Py_OptimizeFlag = add_flag(Py_OptimizeFlag, p);
 
 	interp = PyInterpreterState_New();
 	if (interp == NULL)
@@ -114,6 +127,8 @@ Py_Initialize(void)
 		Py_FatalError("Py_Initialize: can't make first thread");
 	(void) PyThreadState_Swap(tstate);
 
+	_Py_ReadyTypes();
+
 	interp->modules = PyDict_New();
 	if (interp->modules == NULL)
 		Py_FatalError("Py_Initialize: can't make modules dictionary");
@@ -121,8 +136,10 @@ Py_Initialize(void)
 	/* Init codec registry */
 	_PyCodecRegistry_Init();
 
+#ifdef Py_USING_UNICODE
 	/* Init Unicode implementation; relies on the codec registry */
 	_PyUnicode_Init();
+#endif
 
 	bimod = _PyBuiltin_Init();
 	if (bimod == NULL)
@@ -143,7 +160,7 @@ Py_Initialize(void)
 	_PyImport_Init();
 
 	/* initialize builtin exceptions */
-	init_exceptions();
+	_PyExc_Init();
 	_PyImport_FixupExtension("exceptions", "exceptions");
 
 	/* phase 2 of builtins */
@@ -202,9 +219,6 @@ Py_Finalize(void)
 	/* Disable signal handling */
 	PyOS_FiniInterrupts();
 
-	/* Cleanup Unicode implementation */
-	_PyUnicode_Fini();
-
 	/* Cleanup Codec registry */
 	_PyCodecRegistry_Fini();
 
@@ -224,11 +238,7 @@ Py_Finalize(void)
 #endif
 
 #ifdef Py_TRACE_REFS
-	if (
-#ifdef MS_WINDOWS /* Only ask on Windows if env var set */
-	    getenv("PYTHONDUMPREFS") &&
-#endif /* MS_WINDOWS */
-	    _Py_AskYesNo("Print left references?")) {
+	if (Py_GETENV("PYTHONDUMPREFS")) {
 		_Py_PrintReferences(stderr);
 	}
 #endif /* Py_TRACE_REFS */
@@ -238,7 +248,7 @@ Py_Finalize(void)
 	   below has been checked to make sure no exceptions are ever
 	   raised.
 	*/
-	fini_exceptions();
+	_PyExc_Fini();
 
 	/* Delete current thread */
 	PyInterpreterState_Clear(interp);
@@ -253,6 +263,11 @@ Py_Finalize(void)
 	PyInt_Fini();
 	PyFloat_Fini();
 
+#ifdef Py_USING_UNICODE
+	/* Cleanup Unicode implementation */
+	_PyUnicode_Fini();
+#endif
+
 	/* XXX Still allocated:
 	   - various static ad-hoc pointers to interned strings
 	   - int and float free list blocks
@@ -395,8 +410,8 @@ char *
 Py_GetPythonHome(void)
 {
 	char *home = default_home;
-	if (home == NULL)
-		home = getenv("PYTHONHOME");
+	if (home == NULL && !Py_IgnoreEnvironmentFlag)
+		home = Py_GETENV("PYTHONHOME");
 	return home;
 }
 
@@ -495,7 +510,7 @@ PyRun_InteractiveLoopFlags(FILE *fp, char *filename, PyCompilerFlags *flags)
 
 	if (flags == NULL) {
 		flags = &local_flags;
-		local_flags.cf_nested_scopes = 0;
+		local_flags.cf_flags = 0;
 	}
 	v = PySys_GetObject("ps1");
 	if (v == NULL) {
@@ -534,6 +549,7 @@ PyRun_InteractiveOneFlags(FILE *fp, char *filename, PyCompilerFlags *flags)
 	node *n;
 	perrdetail err;
 	char *ps1 = "", *ps2 = "";
+
 	v = PySys_GetObject("ps1");
 	if (v != NULL) {
 		v = PyObject_Str(v);
@@ -550,8 +566,11 @@ PyRun_InteractiveOneFlags(FILE *fp, char *filename, PyCompilerFlags *flags)
 		else if (PyString_Check(w))
 			ps2 = PyString_AsString(w);
 	}
-	n = PyParser_ParseFile(fp, filename, &_PyParser_Grammar,
-			       Py_single_input, ps1, ps2, &err);
+	n = PyParser_ParseFileFlags(fp, filename, &_PyParser_Grammar,
+			    	    Py_single_input, ps1, ps2, &err,
+			    	    (flags &&
+			    	     flags->cf_flags & CO_GENERATOR_ALLOWED) ?
+			    	    	PyPARSE_YIELD_IS_KEYWORD : 0);
 	Py_XDECREF(v);
 	Py_XDECREF(w);
 	if (n == NULL) {
@@ -678,13 +697,19 @@ PyRun_SimpleFileExFlags(FILE *fp, char *filename, int closeit,
 
 int
 PyRun_SimpleString(char *command)
+{
+	return PyRun_SimpleStringFlags(command, NULL);
+}
+
+int
+PyRun_SimpleStringFlags(char *command, PyCompilerFlags *flags)
 {
 	PyObject *m, *d, *v;
 	m = PyImport_AddModule("__main__");
 	if (m == NULL)
 		return -1;
 	d = PyModule_GetDict(m);
-	v = PyRun_String(command, Py_file_input, d, d);
+	v = PyRun_StringFlags(command, Py_file_input, d, d, flags);
 	if (v == NULL) {
 		PyErr_Print();
 		return -1;
@@ -895,7 +920,7 @@ void PyErr_Display(PyObject *exception, PyObject *value, PyObject *tb)
 		if (tb && tb != Py_None)
 			err = PyTraceBack_Print(tb, f);
 		if (err == 0 &&
-		    PyErr_GivenExceptionMatches(exception, PyExc_SyntaxError))
+		    PyObject_HasAttrString(v, "print_file_and_line"))
 		{
 			PyObject *message;
 			char *filename, *text;
@@ -911,7 +936,7 @@ void PyErr_Display(PyObject *exception, PyObject *value, PyObject *tb)
 				else
 					PyFile_WriteString(filename, f);
 				PyFile_WriteString("\", line ", f);
-				sprintf(buf, "%d", lineno);
+				PyOS_snprintf(buf, sizeof(buf), "%d", lineno);
 				PyFile_WriteString(buf, f);
 				PyFile_WriteString("\n", f);
 				if (text != NULL)
@@ -1005,7 +1030,10 @@ PyObject *
 PyRun_StringFlags(char *str, int start, PyObject *globals, PyObject *locals,
 		  PyCompilerFlags *flags)
 {
-	return run_err_node(PyParser_SimpleParseString(str, start),
+	return run_err_node(PyParser_SimpleParseStringFlags(
+			str, start,
+			(flags && flags->cf_flags & CO_GENERATOR_ALLOWED) ?
+				PyPARSE_YIELD_IS_KEYWORD : 0),
 			    "<string>", globals, locals, flags);
 }
 
@@ -1021,7 +1049,9 @@ PyObject *
 PyRun_FileExFlags(FILE *fp, char *filename, int start, PyObject *globals,
 		  PyObject *locals, int closeit, PyCompilerFlags *flags)
 {
-	node *n = PyParser_SimpleParseFile(fp, filename, start);
+	node *n = PyParser_SimpleParseFileFlags(fp, filename, start,
+			(flags && flags->cf_flags & CO_GENERATOR_ALLOWED) ?
+				PyPARSE_YIELD_IS_KEYWORD : 0);
 	if (closeit)
 		fclose(fp);
 	return run_err_node(n, filename, globals, locals, flags);
@@ -1077,12 +1107,8 @@ run_pyc_file(FILE *fp, char *filename, PyObject *globals, PyObject *locals,
 	}
 	co = (PyCodeObject *)v;
 	v = PyEval_EvalCode(co, globals, locals);
-	if (v && flags) {
-		if (co->co_flags & CO_NESTED)
-			flags->cf_nested_scopes = 1;
-		fprintf(stderr, "run_pyc_file: nested_scopes: %d\n",
-			flags->cf_nested_scopes);			
-	}
+	if (v && flags)
+		flags->cf_flags |= (co->co_flags & PyCF_MASK);
 	Py_DECREF(co);
 	return v;
 }
@@ -1099,7 +1125,9 @@ Py_CompileStringFlags(char *str, char *filename, int start,
 {
 	node *n;
 	PyCodeObject *co;
-	n = PyParser_SimpleParseString(str, start);
+	n = PyParser_SimpleParseStringFlags(str, start,
+		(flags && flags->cf_flags & CO_GENERATOR_ALLOWED) ?
+			PyPARSE_YIELD_IS_KEYWORD : 0);
 	if (n == NULL)
 		return NULL;
 	co = PyNode_CompileFlags(n, filename, flags);
@@ -1123,28 +1151,41 @@ Py_SymtableString(char *str, char *filename, int start)
 /* Simplified interface to parsefile -- return node or set exception */
 
 node *
-PyParser_SimpleParseFile(FILE *fp, char *filename, int start)
+PyParser_SimpleParseFileFlags(FILE *fp, char *filename, int start, int flags)
 {
 	node *n;
 	perrdetail err;
-	n = PyParser_ParseFile(fp, filename, &_PyParser_Grammar, start,
-				(char *)0, (char *)0, &err);
+	n = PyParser_ParseFileFlags(fp, filename, &_PyParser_Grammar, start,
+					(char *)0, (char *)0, &err, flags);
 	if (n == NULL)
 		err_input(&err);
 	return n;
 }
 
+node *
+PyParser_SimpleParseFile(FILE *fp, char *filename, int start)
+{
+	return PyParser_SimpleParseFileFlags(fp, filename, start, 0);
+}
+
 /* Simplified interface to parsestring -- return node or set exception */
 
+node *
+PyParser_SimpleParseStringFlags(char *str, int start, int flags)
+{
+	node *n;
+	perrdetail err;
+	n = PyParser_ParseStringFlags(str, &_PyParser_Grammar, start, &err,
+				      flags);
+	if (n == NULL)
+		err_input(&err);
+	return n;
+}
+
 node *
 PyParser_SimpleParseString(char *str, int start)
 {
-	node *n;
-	perrdetail err;
-	n = PyParser_ParseString(str, &_PyParser_Grammar, start, &err);
-	if (n == NULL)
-		err_input(&err);
-	return n;
+	return PyParser_SimpleParseStringFlags(str, start, 0);
 }
 
 /* Set the error appropriate to the given input error code (see errcode.h) */
@@ -1306,6 +1347,9 @@ initsigs(void)
 #ifdef SIGPIPE
 	signal(SIGPIPE, SIG_IGN);
 #endif
+#ifdef SIGXFZ
+	signal(SIGXFZ, SIG_IGN);
+#endif
 #endif /* HAVE_SIGNAL_H */
 	PyOS_InitInterrupts(); /* May imply initsignal() */
 }
@@ -1318,7 +1362,7 @@ _Py_AskYesNo(char *prompt)
 {
 	char buf[256];
 	
-	printf("%s [ny] ", prompt);
+	fprintf(stderr, "%s [ny] ", prompt);
 	if (fgets(buf, sizeof buf, stdin) == NULL)
 		return 0;
 	return buf[0] == 'y' || buf[0] == 'Y';
@@ -1396,6 +1440,12 @@ PyOS_getsig(int sig)
 {
 #ifdef HAVE_SIGACTION
 	struct sigaction context;
+	/* Initialize context.sa_handler to SIG_ERR which makes about as
+	 * much sense as anything else.  It should get overwritten if
+	 * sigaction actually succeeds and otherwise we avoid an
+	 * uninitialized memory read.
+	 */
+	context.sa_handler = SIG_ERR;
 	sigaction(sig, NULL, &context);
 	return context.sa_handler;
 #else
@@ -1412,6 +1462,12 @@ PyOS_setsig(int sig, PyOS_sighandler_t handler)
 #ifdef HAVE_SIGACTION
 	struct sigaction context;
 	PyOS_sighandler_t oldhandler;
+	/* Initialize context.sa_handler to SIG_ERR which makes about as
+	 * much sense as anything else.  It should get overwritten if
+	 * sigaction actually succeeds and otherwise we avoid an
+	 * uninitialized memory read.
+	 */
+	context.sa_handler = SIG_ERR;
 	sigaction(sig, NULL, &context);
 	oldhandler = context.sa_handler;
 	context.sa_handler = handler;
