@@ -3,6 +3,7 @@ from bgenGeneratorGroup import GeneratorGroup
 
 class ObjectDefinition(GeneratorGroup):
 	"Spit out code that together defines a new Python object type"
+	basechain = "NULL"
 
 	def __init__(self, name, prefix, itselftype):
 		"""ObjectDefinition constructor.  May be extended, but do not override.
@@ -21,7 +22,7 @@ class ObjectDefinition(GeneratorGroup):
 		self.typename = name + '_Type'
 		self.argref = ""	# set to "*" if arg to <type>_New should be pointer
 		self.static = "static " # set to "" to make <type>_New and <type>_Convert public
-		self.basechain = "NULL" # set to &<basetype>_chain to chain methods
+		self.modulename = None
 
 	def add(self, g, dupcheck=0):
 		g.setselftype(self.objecttype, self.itselftype)
@@ -30,6 +31,9 @@ class ObjectDefinition(GeneratorGroup):
 	def reference(self):
 		# In case we are referenced from a module
 		pass
+		
+	def setmodulename(self, name):
+		self.modulename = name
 
 	def generate(self):
 		# XXX This should use long strings and %(varname)s substitution!
@@ -80,10 +84,8 @@ class ObjectDefinition(GeneratorGroup):
 
 	def outputNew(self):
 		Output()
-		Output("%sPyObject *%s_New(itself)", self.static, self.prefix)
-		IndentLevel()
-		Output("%s %sitself;", self.itselftype, self.argref)
-		DedentLevel()
+		Output("%sPyObject *%s_New(%s %sitself)", self.static, self.prefix,
+				self.itselftype, self.argref)
 		OutLbrace()
 		Output("%s *it;", self.objecttype)
 		self.outputCheckNewArg()
@@ -100,11 +102,8 @@ class ObjectDefinition(GeneratorGroup):
 			"Override this method to apply additional checks/conversions"
 	
 	def outputConvert(self):
-		Output("%s%s_Convert(v, p_itself)", self.static, self.prefix)
-		IndentLevel()
-		Output("PyObject *v;")
-		Output("%s *p_itself;", self.itselftype)
-		DedentLevel()
+		Output("%sint %s_Convert(PyObject *v, %s *p_itself)", self.static, self.prefix,
+				self.itselftype)
 		OutLbrace()
 		self.outputCheckConvertArg()
 		Output("if (!%s_Check(v))", self.prefix)
@@ -121,10 +120,7 @@ class ObjectDefinition(GeneratorGroup):
 
 	def outputDealloc(self):
 		Output()
-		Output("static void %s_dealloc(self)", self.prefix)
-		IndentLevel()
-		Output("%s *self;", self.objecttype)
-		DedentLevel()
+		Output("static void %s_dealloc(%s *self)", self.prefix, self.objecttype)
 		OutLbrace()
 		self.outputCleanupStructMembers()
 		Output("PyMem_DEL(self);")
@@ -138,11 +134,7 @@ class ObjectDefinition(GeneratorGroup):
 
 	def outputGetattr(self):
 		Output()
-		Output("static PyObject *%s_getattr(self, name)", self.prefix)
-		IndentLevel()
-		Output("%s *self;", self.objecttype)
-		Output("char *name;")
-		DedentLevel()
+		Output("static PyObject *%s_getattr(%s *self, char *name)", self.prefix, self.objecttype)
 		OutLbrace()
 		self.outputGetattrBody()
 		OutRbrace()
@@ -176,9 +168,12 @@ class ObjectDefinition(GeneratorGroup):
 		Output()
 		Output("%sPyTypeObject %s = {", sf, self.typename)
 		IndentLevel()
-		Output("PyObject_HEAD_INIT(&PyType_Type)")
+		Output("PyObject_HEAD_INIT(NULL)")
 		Output("0, /*ob_size*/")
-		Output("\"%s\", /*tp_name*/", self.name)
+		if self.modulename:
+			Output("\"%s.%s\", /*tp_name*/", self.modulename, self.name)
+		else:
+			Output("\"%s\", /*tp_name*/", self.name)
 		Output("sizeof(%s), /*tp_basicsize*/", self.objecttype)
 		Output("0, /*tp_itemsize*/")
 		Output("/* methods */")
@@ -201,7 +196,7 @@ class ObjectDefinition(GeneratorGroup):
 		Output("""if (PyDict_SetItemString(d, "%sType", (PyObject *)&%s) != 0)""",
 			self.name, self.typename);
 		IndentLevel()
-		Output("""Py_FatalError("can't initialize %sType");""",
+		Output("""Py_FatalError("can\'t initialize %sType");""",
 		                                           self.name)
 		DedentLevel()
 
@@ -226,10 +221,8 @@ class ObjectIdentityMixin:
 	
 	def outputCompare(self):
 		Output()
-		Output("static int %s_compare(self, other)", self.prefix)
-		IndentLevel()
-		Output("%s *self, *other;", self.objecttype)
-		DedentLevel()
+		Output("static int %s_compare(%s *self, %s *other)", self.prefix, self.objecttype,
+				self.objecttype)
 		OutLbrace()
 		Output("unsigned long v, w;")
 		Output()
@@ -250,10 +243,7 @@ class ObjectIdentityMixin:
 		
 	def outputHash(self):
 		Output()
-		Output("static long %s_hash(self)", self.prefix)
-		IndentLevel()
-		Output("%s *self;", self.objecttype)
-		DedentLevel()
+		Output("static long %s_hash(%s *self)", self.prefix, self.objecttype)
 		OutLbrace()
 		Output("return (long)self->ob_itself;")
 		OutRbrace()
